/**
 * Copyright (c) 2011, REMPL.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1) Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2) Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following
 * disclaimer in the documentation and/or other materials provided
 * with the distribution.
 * 3) Neither the name of the REMPL.com nor the names of its
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.rempl.core;

// REMPL API from com.rempl:rempl-api
import com.rempl.api.Element;
import com.rempl.api.Logger;
import com.rempl.api.Model;

// for slots manipulation
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang.StringUtils;

/**
 * MOF 2.0 Element (UML Element actually).
 *
 * @author Yegor Bugayenko (yegor@rempl.com)
 * @version $Id: ElementImpl.java 1524 2011-06-27 17:41:05Z guard $
 */
public final class ElementImpl implements Element {

    /**
     * Multi-valued map of slots. The variable is NOT
     * final since its value is changes in {@link #renameRefs(String,String)}.
     * @see #renameRefs(String,String)
     */
    private Map<String, List<String>> slots =
        new HashMap<String, List<String>>();

    /**
     * This element is ready for the model or it is still
     * under preparation and parsing?
     * @see #validate(String)
     */
    private boolean valid;

    /**
     * ID of the element.
     */
    private String name;

    /**
     * Name of MOF meta-class. For example it could be "Class"
     * or "Package", or "File".
     */
    private String type;

    /**
     * Public ctor.
     * @param pname The id of the element
     * @param ptype The name of the meta-class.
     */
    public ElementImpl(final String pname, final String ptype) {
        this.name = pname;
        this.type = ptype;
    }

    /**
     * Rename the element.
     * @param pname New name of the element
     * @see SimpleModel#rename(Collection)
     */
    protected void rename(final String pname) {
        if (!pname.matches("^[a-z]\\d+$")) {
            throw new IllegalArgumentException(
                "Element can't be renamed to \""
                + pname
                + "\", but only to something that"
                + " matches \"[a-z]\\d+\" pattern"
            );
        }
        this.name = pname;
        this.valid = true;
    }

    /**
     * Rename all refs inside this element.
     * @param old Old name of reference
     * @param update New name to use for all these references
     * @see SimpleModel#rename(Collection)
     * @todo #18 I don't like this implementation. It's too complex and
     *       very likely is slow. Something more compact and effective
     *       should be implemented here instead.
     */
    protected void renameRefs(final String old, final String update) {
        final Map<String, List<String>> newSlots =
            new HashMap<String, List<String>>();
        for (Map.Entry<String, List<String>> entry : this.slots.entrySet()) {
            final List<String> values = new ArrayList<String>();
            for (String value : entry.getValue()) {
                String newValue = value;
                if (StringUtils.startsWith(value, Model.REF_PREFIX)) {
                    final String ref = value.substring(1);
                    if (ref.equals(old)) {
                        newValue = Model.REF_PREFIX + update;
                        Logger.debug(
                            this,
                            "#renameRefs('%s', '%s'): slot '%s:%s'"
                            + " renamed to '%s:%s' in %s",
                            old,
                            update,
                            entry.getKey(),
                            value,
                            entry.getKey(),
                            newValue,
                            this.name
                        );
                    }
                }
                values.add(newValue);
            }
            newSlots.put(entry.getKey(), values);
        }
        this.slots = newSlots;
    }

    /**
     * Get its number.
     * @return The number of the element (ordered in the model)
     * @see SimpleModel#rename(Collection)
     */
    protected Integer number() {
        if (!this.valid) {
            throw new IllegalStateException(
                "Element \""
                + this.name
                + "\" is not valid yet, can't get its number"
            );
        }
        return Integer.valueOf(this.name.substring(1));
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String name() {
        return this.name;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String type() {
        return this.type;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean has(final String key) {
        return this.slots.containsKey(key);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String get(final String key) {
        if (!this.has(key)) {
            throw new IllegalArgumentException(
                "Invalid key in "
                + this.name
                + " (" + this.type + ")"
            );
        }
        return this.slots.get(key).get(0);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public <T> T get(final String key, final Class<T> cls) {
        T result;
        if (cls.equals(String.class)) {
            result = (T) this.get(key);
        } else if (cls.equals(Integer.class)) {
            result = (T) Integer.valueOf(this.get(key));
        } else if (cls.equals(Boolean.class)) {
            if ("TRUE".equals(this.get(key))) {
                result = (T) Boolean.TRUE;
            } else {
                result = (T) Boolean.FALSE;
            }
        } else {
            throw new IllegalArgumentException(
                String.format(
                    "get(%s, %s) is not implemented yet",
                    key,
                    cls.getName()
                )
            );
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean refsTo(final String key, final Element another) {
        if (!this.slots.containsKey(key)) {
            throw new IllegalArgumentException(
                String.format(
                    "Slot \"%s\" is absent in %s:%s",
                    key,
                    this.name,
                    this.type
                )
            );
        }
        if (another == null) {
            throw new IllegalArgumentException(
                "Another element can't be NULL"
            );
        }
        if (key == null) {
            throw new IllegalArgumentException(
                "Name of the slot (key) can't be NULL"
            );
        }
        for (String value : this.slots.get(key)) {
            if (value.equals(Model.REF_PREFIX + another.name())) {
                return true;
            }
        }
        return false;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void put(final String key, final String value) {
        if (!this.has(key)) {
            this.slots.put(key, new ArrayList<String>());
        }
        this.slots.get(key).add(value);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Collection<String> all(final String key) {
        return this.slots.get(key);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Collection<Map.Entry<String, String>> entrySet() {
        final Collection<Map.Entry<String, String>> list =
            new ArrayList<Map.Entry<String, String>>();
        for (Map.Entry<String, List<String>> entry : this.slots.entrySet()) {
            for (String value : entry.getValue()) {
                list.add(
                    new HashMap.SimpleEntry<String, String>(
                        entry.getKey(),
                        value
                    )
                );
            }
        }
        return list;
    }

}
