/**
 * Copyright (c) 2011, REMPL.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1) Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2) Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following
 * disclaimer in the documentation and/or other materials provided
 * with the distribution.
 * 3) Neither the name of the REMPL.com nor the names of its
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.rempl.core;

// API classes
import com.rempl.api.Element;
import com.rempl.api.Logger;
import com.rempl.api.Model;
import com.rempl.api.RuntimeProblem;

// for collection implementation
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

// IO utility class, and string manipulation
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.StringUtils;

/**
 * MOF 2.0 model abstraction.
 *
 * @author Yegor Bugayenko (yegor@rempl.com)
 * @version $Id: SimpleModel.java 1524 2011-06-27 17:41:05Z guard $
 * @todo #2 This implementation is extremely simplified and won't
 *          be able to do anything good in any real-life project.
 *          It has to be extended significantly or replaced at all
 *          by Eclipse EMF components.
 * @todo #2:0.5h More INFO-level reporting is required
 */
public final class SimpleModel implements Model {

    /**
     * List of elements.
     */
    private final Collection<ElementImpl> elements =
        new ArrayList<ElementImpl>();

    /**
     * {@inheritDoc}
     */
    @Override
    public void transform(final String query) {
        final Collection<ElementImpl> found =
            new TransformAnalyzer().analyze(query);
        this.validate(found);
        this.rename(found);
        for (ElementImpl e : found) {
            this.elements.add(e);
        }
        Logger.debug(
            this,
            "#transform(%s): %d elements imported",
            query,
            found.size()
        );
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Collection<Element> query(final String query) {
        final String context = new QueryAnalyzer().analyze(query);
        final Collection<Element> list = new ArrayList<Element>();
        for (Element e : this.elements) {
            if (e.type().equals(context)) {
                list.add(e);
            }
        }
        Logger.debug(
            this,
            "#query('%s'): found %d elements",
            query,
            list.size()
        );
        return list;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void save(final java.io.Writer writer) throws java.io.IOException {
        final StringBuffer buffer = new StringBuffer();
        for (Element e : this.elements) {
            if (e.name() != null) {
                buffer.append(e.name());
            }
            buffer.append(":" + e.type() + " { ");
            for (Map.Entry<String, String> pair : e.entrySet()) {
                buffer.append(
                    pair.getKey()
                    + ": \""
                    + pair.getValue()
                    + "\"; "
                );
            }
            buffer.append(" };\n");
        }
        writer.write(buffer.toString());
        Logger.debug(
            this,
            "#save(%s): saved %d bytes",
            writer.getClass().getName(),
            buffer.length()
        );
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void load(final java.io.Reader reader) throws java.io.IOException {
        final String xmi = IOUtils.toString(reader);
        this.transform(xmi);
        Logger.debug(
            this,
            "#load(%s): loaded %d bytes",
            reader.getClass().getName(),
            xmi.length()
        );
    }

    /**
     * Check the list of elements, for validity.
     * @param list List of elements in the transformation
     * @see #transform(String)
     */
    private void validate(final Collection<ElementImpl> list) {
        final Set<String> names = new HashSet<String>();
        // find possible duplicates
        for (Element e : list) {
            if (names.contains(e.name())) {
                throw RuntimeProblem.make("Duplicate name '%s'", e.name());
            }
            names.add(e.name());
        }
        // make sure all links/refs are valid inside this sub-model

        for (ElementImpl e : list) {
            for (Map.Entry<String, String> entry : e.entrySet()) {
                if (!StringUtils.startsWith(entry.getValue(),
                            Model.REF_PREFIX)) {
                    continue;
                }
                final String ref = entry.getValue().substring(1);
                if (!names.contains(ref)) {
                    throw RuntimeProblem.make(
                        "Invalid reference '%s.%s'",
                        e.name(),
                        ref
                    );
                }
            }
        }
        Logger.debug(
            this,
            "#validate(%d elements): valid names (%s)",
            list.size(),
            StringUtils.join(names, ", ")
        );
    }

    /**
     * Rename all elements from the list, in order to make
     * their names compliant with our internal holder.
     * @param list List of elements in the transformation
     * @see #transform(String)
     * @todo #18:3h This implementation is not performance-effective
     *       at all and has to be replaced with some sort of
     *       recursion. I don't know what exactly will work here, but
     *       current solution is definitively is ineffective.
     */
    private void rename(final Collection<ElementImpl> list) {
        Integer number = 0;
        for (ElementImpl e : this.elements) {
            if (e.number() > number) {
                number = e.number();
            }
        }
        for (ElementImpl e : list) {
            number += 1;
            final String newName = "e" + number;
            for (ElementImpl other : list) {
                other.renameRefs(e.name(), newName);
            }
            e.rename(newName);
        }
        Logger.debug(
            this,
            "#rename(%d elements): renamed, new maximum is %d",
            list.size(),
            number
        );
    }

}
